#!/usr/bin/env python3
import random, binascii, os, gmpy2
random = random.SystemRandom()
from Crypto.Hash import SHA256 as SHA

p = int('7979797979797900ffffffffffffffff66f1ab3f3086177566aeeb771737a3bc'
        '5d73515634204ecbe4b840d457843b454f89764df095d7d762a3a9eed1642e4b', 16)
pbytes = p.bit_length() + 7 >> 3
keylen = 16
threshold = 2
num_shares = 3
tobytes = lambda x: int(x).to_bytes(pbytes, 'little')
frombytes = lambda x: Fp(int.from_bytes(x, 'little'))
tohex = lambda bs: binascii.hexlify(bs).decode()
fromhex = lambda s: binascii.unhexlify(s)

# just integers modulo p, nothing to see here
class Fp:
    def __init__(self, val = 0): self.val = int(val) % p
    def __add__(self, other): return type(self)(self.val + int(other))
    def __sub__(self, other): return type(self)(self.val - int(other))
    def __mul__(self, other): return type(self)(self.val * int(other))
    def __pow__(self, e): return type(self)(pow(self.val, e, p))
    def __truediv__(self, other): return self * gmpy2.invert(int(other), p)
    def __str__(self): return hex(self.val)
    __repr__ = __str__
    def __eq__(self, other): return self.val == other.val
    def __int__(self): return self.val
    __hash__ = __int__

def split(secret, thresh, numsh):
    coeffs = [secret] + [Fp(random.randrange(p)) for _ in range(1, thresh)]
    shares = []
    while len(shares) < numsh:
        x = Fp(random.randrange(1, p))
        y = sum((c * x ** e for e, c in enumerate(coeffs)), Fp())
        shares.append((x, y))
    return shares

def combine(shares, at = Fp()):
    assert len(set(i for i, _ in shares)) == len(shares)
    # Neville's algorithm
    js, rs = zip(*shares)
    for k in range(1, len(rs)):
        rs = [((j - at) * l + (at - i) * r) / (j - i) \
                for i, j, l, r in zip(js[:-k], js[k:], rs, rs[1:])]
    return rs[0]

def pack(msg, key = b''):
    if isinstance(msg, str): msg = msg.encode()
    return SHA.new(key + msg).digest() + msg

def unpack(msg, key = b''):
    tag = msg[:SHA.digest_size]
    if tag == SHA.new(key + msg[len(tag):]).digest():
        return msg[len(tag):]
    print('bad {}: {}'.format(('checksum', 'tag')[bool(key)], tohex(tag)))

################################################################################

if __name__ == '__main__':

    key = os.urandom(keylen)

    # hand over a fair share
    packed = Fp(frombytes(pack(r'echo Hello\!;#' + tohex(os.urandom(32)))))
    shares = split(packed, threshold, num_shares)
    share = shares.pop()
    authed = pack(b''.join(tobytes(s) for s in share), key)
    print('remember this: {}'.format(binascii.hexlify(authed).decode()))

    for _ in range(0x123):

        print('what have you got?')

        inp = unpack(binascii.unhexlify(input()), key)
        if inp is None:
            continue

        share = tuple(map(frombytes, (inp[:pbytes], inp[pbytes:])))
        combined = combine([share] + shares[1:threshold])
        cmd = unpack(tobytes(combined).rstrip(b'\0'))
        if cmd is None:
            continue

        os.system(cmd.decode())

